@@ -1,5 +1,5 @@
 /*  lmpc -- the Little Movie Processing Centre
-    Copyright (C) 1994-99 Uwe Girlich
+    Copyright (C) 1994-2000 Uwe Girlich
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -544,12 +544,29 @@ void WriteOffsetVec(BB_t* m, float* a)
   for ( i=0 ; i<3 ; i++ ) WriteOffset(m, a[i] );
 }
 
-void do_simple_message_write_bin(node* n, BB_t* m)
+
+void
+do_simple_message_write_bin(node* n, BB_t* m)
+{
+  unicast_t u;
+  
+  u.use = UC_NOT;
+  do_simple_message_write_bin_unicast(n,m,&u);
+}
+
+
+void
+do_simple_message_write_bin_unicast(node* n, BB_t* m, unicast_t* u)
 {
   node* c;
- 
+
   /* at first: the message id */
-  WriteByte(m,node_token_id(n->type));
+  if (u->use == UC_THIS) {
+    WriteByte(m,node_token_id(n->type) | MASK_UNICAST);
+    WriteByte(m,u->client);
+  }
+  else
+    WriteByte(m,node_token_id(n->type));
 
   if ((c = n->down) != NULL) {
     /* c is the first command or the first argument */
@@ -577,7 +594,7 @@ void do_unknown_message_write_bin(node* n, BB_t* m)
     I don't do any kind of check in here. If the internal structure is
     corrupt this gives a total crash.
 
-    unknwon, message id itself is an argument
+    unknown, message id itself is an argument
   */
 
   /* message -> commands */
