@@ -1,5 +1,5 @@
 /*  lmpc -- the Little Movie Processing Centre
-    Copyright (C) 1994-99 Uwe Girlich
+    Copyright (C) 1994-2000 Uwe Girlich
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -647,7 +647,7 @@ int main(int argc, char **argv)
 void syntax(void)
 {
   fprintf(stderr,"Little Movie Processing Centre\n");
-  fprintf(stderr,"LMPC (c) U. Girlich, 1994-99, Release %s %s (%s)\n", VERSION, DATE, COMMENT);
+  fprintf(stderr,"LMPC (c) U. Girlich, 1994-2000, Release %s %s (%s)\n", VERSION, DATE, COMMENT);
   fprintf(stderr,"lmpc [option ...] filename [filename ...]\n");
   fprintf(stderr,"-h, --help                   display this help and exit.\n");
   fprintf(stderr,"-V, --version                output version information and exit.\n");
@@ -2947,7 +2947,7 @@ void ActionDM2bin2DM2bin(char *dm2binfilename1, char *dm2binfilename2,
     DM2_block_read_bin(&d1, &m);
     if (m.size > DM2_MAX_MESSAGE_SIZE) {
       /* clean the record type hint */
-      m.hint &= ~(DM2_RECORD_CLIENT|DM2_RECORD_SERVER);
+      m.hint &= ~(DM2_RECORD_CLIENT|DM2_RECORD_SERVER|DM2_RECORD_RELAY);
       /* set serverrecord variant */
       m.hint |= DM2_RECORD_SERVER;
     }
@@ -2996,7 +2996,7 @@ void ActionDM2bin2DM2txt(char *dm2binfilename, char *dm2txtfilename,
     DM2_block_read_bin(&d, &m);
     if (m.size > DM2_MAX_MESSAGE_SIZE) {
       /* clean the record type hint */
-      m.hint &= ~(DM2_RECORD_CLIENT|DM2_RECORD_SERVER);
+      m.hint &= ~(DM2_RECORD_CLIENT|DM2_RECORD_SERVER|DM2_RECORD_RELAY);
       /* set serverrecord variant */
       m.hint |= DM2_RECORD_SERVER;
     }
@@ -3092,7 +3092,7 @@ void ActionDM2txt2DM2txt(char *dm2txtfilename1, char *dm2txtfilename2,
 
 
 /*---------------------------------------------------------------------------*/
-/* DM2 */
+/* DM3 */
 
 
 #ifdef ENABLE_DM3
@@ -3129,7 +3129,7 @@ void ActionDM3bin2DM3txt(char *dm3binfilename, char *dm3txtfilename,
   DM3_t d;
   TEXT_t s;
   char ts[1000];
-  BB_t m;
+  DM3_binblock m;
   node *n;
   int o;
 
