@@ -157,6 +157,7 @@ node* node_combine3(node* n1, node* n2, node* n3)
 
 node* node_add_comment(node* n, char* comment)
 {
+char a;
   /* do only anything, if the comment is not NULL */
   if (comment != NULL) {
     /* the latest survives, I should maybe concatenate both strings ?! */
@@ -165,6 +166,8 @@ node* node_add_comment(node* n, char* comment)
       n->comment=NULL;
     }
     n->comment=comment;
+    a = n->comment[0];
+    /* fprintf(stderr,"n=0x%08lx *0x%08lx=0x%02x\n",(long)n,(long)n->comment,a); */
   }
   return n;
 }
