@@ -1,5 +1,5 @@
 /*  lmpc -- the Little Movie Processing Centre
-    Copyright (C) 1994-99 Uwe Girlich
+    Copyright (C) 1994-2000 Uwe Girlich
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -182,7 +182,9 @@ token_t DM2_token[]={
 	{ "pos3",			TOKEN_POS3,			0x00 },
 	{ "pos4",			TOKEN_POS4,			0x00 },
 	{ "plat2flags",			TOKEN_PLAT2FLAGS,		0x00 },
-	{ "wait",			TOKEN_WAIT,			0x00 }
+	{ "wait",			TOKEN_WAIT,			0x00 },
+	{ "connected",			TOKEN_CONNECTED,		0x00 },
+	{ "unicast",			TOKEN_UNICAST,			0x00 }
 };
 
 void udm2_init(void)
@@ -305,7 +307,7 @@ void DM2_init(DM2_t* d, char *filename, char *mode)
   /* we need a rewind here because some fopen modes may seek at the end */
   rewind(d->file);
   d->serverversion = 26; /* standard setting, retail CD version 3.05 */
-  d->demotype = DM2_RECORD_CLIENT|DM2_RECORD_SERVER; /* both possible ways */
+  d->demotype = DM2_RECORD_CLIENT|DM2_RECORD_SERVER|DM2_RECORD_RELAY; /* all possible ways */
 }
 
 void DM2_prepare_write_bin(DM2_t* d)
@@ -379,16 +381,15 @@ node* DM2_bin_to_node(BB_t *m, int opt)
   unsigned long size;
   char ts[1000];
   long id;
+  node *messagenode;
+  node *blocknode;
 
-  n=NULL;
   size = ReadLong(m); 
   if (size == 0) {
-    n = node_init(TOKEN_BETWEENBLOCK,NULL,0);
-    return n;
+    return node_init(TOKEN_BETWEENBLOCK,NULL,0);
   }
   if (size == 0xFFFFFFFF) {
-    n = node_init(TOKEN_ENDBLOCK,NULL,0);
-    return n;
+    return node_init(TOKEN_ENDBLOCK,NULL,0);
   }
 
   /*
@@ -396,16 +397,31 @@ node* DM2_bin_to_node(BB_t *m, int opt)
                   m->base + (m->p-m->buffer),
                   m->base + (m->p-m->buffer) + size - 1);
   */
+  messagenode=NULL;
   while (m->p < m->end) {
-    /* fprintf(stderr, "pos=0x%08lX\n", m->base + (m->p-m->buffer)); */ 
+    /* fprintf(stderr, "pos=0x%08lX\n", m->base + (m->p-m->buffer)); */
     if (opt & 0x01) {
       tn=node_init_all(GEN_NOTHING,H_NOTHING,NULL,0);
       sprintf(ts, "0x%08lX", m->base + (m->p-m->buffer));
       node_add_comment(tn,NODE_VALUE_STRING_dup(ts));
-      n=node_link(n,tn);
+      messagenode=node_link(messagenode,tn);
     }
     id = ReadByte(m);
     /* fprintf(stderr,"id=0x%02lX\n", id); */ 
+    if (id & MASK_UNICAST) {
+      long unicast;
+      node* unicastnode;
+
+      unicast = ReadByte(m);
+      id &= ~MASK_UNICAST;
+      unicastnode = node_add_hint(
+                      node_command_init(TOKEN_UNICAST, V_INT, H_BYTE,
+                                        NODE_VALUE_INT_dup(unicast), 0),
+                      H_UNICAST);
+                                     
+      messagenode = node_link(messagenode, unicastnode);
+    }
+    n=NULL;
     switch(id) {
       case 0x00: { /* bad */
         /* node tree */
@@ -1361,8 +1377,10 @@ node* DM2_bin_to_node(BB_t *m, int opt)
         unsigned char* p_base;
         unsigned char* p_client = 0;
         unsigned char* p_server = 0;
+        unsigned char* p_relay  = 0;
         unsigned char id_client;
         unsigned char id_server;
+        unsigned char id_relay;
         int rectype;
               
         int i;
@@ -1371,20 +1389,22 @@ node* DM2_bin_to_node(BB_t *m, int opt)
         long uk_b1 = 0;
         long count = 0;
         long areas[32];
+        long connected_count = 0;
+        long connected[DM2_MAX_CLIENTS];
         long frame = 0;
-        
+
         p_base = m->p; /*base position */
-        /* both is possible and DEMOTYPE reduces it again */
-        rectype = (DM2_RECORD_CLIENT|DM2_RECORD_SERVER) & DEMOTYPE;
+        /* all is possible and DEMOTYPE reduces it again */
+        rectype = (DM2_RECORD_CLIENT|DM2_RECORD_SERVER|DM2_RECORD_RELAY) & DEMOTYPE;
 
         /* nothing at all */
         if (!rectype) 
           syserror(WDM2,DM2TOP->filename);
 
-        /* binary in: client */
 	if (rectype & DM2_RECORD_CLIENT) {
 	  /* client recording is possible */
           m->p = p_base; /* back on base position */
+          /* binary in: client */
           seq1 = ReadLong(m);
           seq2 = ReadLong(m);
           if (SERVERVERSION != 26) {
@@ -1407,6 +1427,7 @@ node* DM2_bin_to_node(BB_t *m, int opt)
         }
 
         if (rectype & DM2_RECORD_SERVER) {
+          /* server recording is possible */
           m->p = p_base; /* back on base position */
           /* binary in: server */
           frame = ReadLong(m);
@@ -1417,9 +1438,43 @@ node* DM2_bin_to_node(BB_t *m, int opt)
           }
         }
 
+	if (rectype & DM2_RECORD_RELAY) {
+	  /* relay recording is possible */
+          m->p = p_base; /* back on base position */
+          /* binary in: relay */
+          seq1 = ReadLong(m);
+          seq2 = ReadLong(m);
+          if (SERVERVERSION != 26) {
+            uk_b1 = ReadByte(m);
+          }
+          count = ReadByte(m);
+          if (count > 32) {
+            rectype &= ~DM2_RECORD_RELAY;
+          }
+          else {
+            for (i=0;i<count;i++) {
+              areas[i] = ReadByte(m);
+            }
+          }
+          connected_count = ReadByte(m);
+          if (connected_count > 32) {
+            rectype &= ~DM2_RECORD_RELAY;
+          }
+          else {
+            for (i=0;i<connected_count;i++) {
+              connected[i] = ReadByte(m);
+            }
+          }
+          p_relay = m->p; /* after relay */
+          id_relay = ReadByte(m) & ~MASK_UNICAST;
+          if (id_relay != 0x11 && id_relay != 0x12) { /* playerinfo, packetentities */
+            rectype &= ~DM2_RECORD_RELAY;
+          }
+        }
+
         /* node tree */
         tn = NULL;
-        /* Now comes the tricky bit. There are 4 possible cases for rectype */
+        /* Now comes the tricky bit. There are many possible cases for rectype */
         switch (rectype) {
           case DM2_RECORD_CLIENT:                    /* sure a client */
           case DM2_RECORD_CLIENT|DM2_RECORD_SERVER:  /* may be both, I choose client */
@@ -1447,11 +1502,39 @@ node* DM2_bin_to_node(BB_t *m, int opt)
           {
             /* point after server parsed message */
             m->p = p_server;
+            
             n = node_link(n,node_add_hint(node_command_init(TOKEN_FRAME,V_INT,H_LONG,
                             NODE_VALUE_INT_dup(frame),0),H_SIMPLE));                  
           }
           break;
-          case 0: /* both record types are wrong, a wrong DM2 file */
+          case DM2_RECORD_RELAY:
+          {
+            /* point after relay parsed message */
+            m->p = p_relay;
+            
+            tn = node_link(tn,node_command_init(TOKEN_SEQ1,V_INT,H_LONG,
+                              NODE_VALUE_INT_dup(seq1),0));
+            tn = node_link(tn,node_command_init(TOKEN_SEQ2,V_INT,H_LONG,
+                              NODE_VALUE_INT_dup(seq2),0));
+            if (SERVERVERSION != 26) {
+              tn = node_link(tn,node_command_init(TOKEN_UK_B1, V_INT, H_BYTE, NODE_VALUE_INT_dup(uk_b1),0));
+            }
+            ttn = NULL;
+            for (i=0;i<count;i++) {
+              ttn=node_link(ttn,node_init_all(V_BYTEHEX,H_SIMPLE,
+                                NODE_VALUE_BYTEHEX_dup(areas[i]),0));
+            }
+            tn = node_link(tn, node_init_all(TOKEN_AREAS, H_SIMPLE, ttn, 0));
+            ttn = NULL;
+            for (i=0;i<connected_count;i++) {
+              ttn=node_link(ttn,node_init_all(V_BYTEHEX,H_SIMPLE,
+                                NODE_VALUE_BYTEHEX_dup(connected[i]),0));
+            }
+            tn = node_link(tn, node_init_all(TOKEN_CONNECTED, H_SIMPLE, ttn, 0));
+            n=node_link(n,node_init_all(TOKEN_FRAME, H_DM2_FRAME, tn, 0));
+          }
+          break;
+          case 0: /* all record types are wrong, a wrong DM2 file */
           {
             syserror(WDM2,DM2TOP->filename);
           }
@@ -1483,13 +1566,14 @@ node* DM2_bin_to_node(BB_t *m, int opt)
       } /* end default */
       break;
     } /* end switch */
+    messagenode = node_link(messagenode,n);
   } /* end while */
-  n = node_init_all(TOKEN_BLOCK,0,n,0);
+  blocknode = node_init_all(TOKEN_BLOCK,0,messagenode,0);
   if (opt & 0x01) {
     sprintf(ts,"%i",DM2TOP->frame);
-    node_add_comment(n,NODE_VALUE_STRING_dup(ts));
-  }       
-  return n;
+    node_add_comment(blocknode,NODE_VALUE_STRING_dup(ts));
+  }
+  return blocknode;
 }
 
 /*----------------------------------------------------------------------------*/
@@ -1522,6 +1606,7 @@ node* DM2_block_write_bin(node* b)
   BB_t m;
   node* tn;
   long datasize = 0;
+  unicast_t unicast;
 
   /* b may be NULL (something was wrong or removed) */
   if (b == NULL) return b;
@@ -1557,36 +1642,45 @@ node* DM2_block_write_bin(node* b)
         for (; tn!=NULL; tn=tn->next) {
           /* do something useful with the tn */
           /* fprintf(stderr,"m %d, h %d\n", tn->type, tn->hint); */
+          if (unicast.use == UC_NEXT) 
+            unicast.use = UC_THIS;
           switch(tn->hint) {
+            case H_UNICAST:
+              unicast.client = *(int*)tn->down->down;
+              unicast.use = UC_NEXT;
+            break;
             case H_NOTHING:
             break;
             case H_SIMPLE: 
-              do_simple_message_write_bin(tn,&m);
+              do_simple_message_write_bin_unicast(tn,&m,&unicast);
             break;
             case H_UNKNOWN:
+              /* TODO: unicast does not work here */
               do_unknown_message_write_bin(tn,&m);
             break;
             case H_DM2_SOUND:
-              do_dm2_sound_message_write_bin(tn,&m);
+              do_dm2_sound_message_write_bin(tn,&m,&unicast);
             break;
             case H_DM2_PLAYERINFO:
-              do_dm2_playerinfo_message_write_bin(tn,&m);
+              do_dm2_playerinfo_message_write_bin(tn,&m,&unicast);
             break;
             case H_DM2_SPAWNBASELINE:
-              do_dm2_spawnbaseline_message_write_bin(tn,&m);
+              do_dm2_spawnbaseline_message_write_bin(tn,&m,&unicast);
             break;
             case H_DM2_PACKETENTITIES:
-              do_dm2_packetentities_message_write_bin(tn,&m);		
+              do_dm2_packetentities_message_write_bin(tn,&m,&unicast);		
             break;
-            case H_DM2_FRAME: /* only client side frame */
-              do_dm2_frame_message_write_bin(tn,&m);
+            case H_DM2_FRAME:
+              do_dm2_frame_message_write_bin(tn,&m,&unicast);
             break;
             default: /* this creates a bad message */
               *m.p='\0';
               m.p++;
             break;
-          }
-        }
+          } /* end switch */
+          if (unicast.use == UC_THIS) 
+            unicast.use = UC_NOT;
+        } /* end while */
         /* calculate the size info */
         datasize = m.p - m.buffer; 
         m.size = (datasize) - 4;
@@ -1615,7 +1709,8 @@ node* DM2_block_write_bin(node* b)
 #define NEXTRET(x) { NEXT(x) ; CHECKRET(x); }
 #define NEXTGOTO(x,g) { NEXT(x) ; if ((x)==NULL) { goto g; } }
 
-void do_dm2_sound_message_write_bin(node* n, BB_t* m)
+void
+do_dm2_sound_message_write_bin(node* n, BB_t* m, unicast_t* u)
 {
   node* cn;
   unsigned char mask;
@@ -1626,7 +1721,12 @@ void do_dm2_sound_message_write_bin(node* n, BB_t* m)
   */
 
   /* at first: the message id */
-  WriteByte(m,node_token_id(n->type));
+  if (u->use == UC_THIS) {
+    WriteByte(m,node_token_id(n->type) | MASK_UNICAST);
+    WriteByte(m,u->client);
+  }
+  else
+    WriteByte(m,node_token_id(n->type));
 
   /* collect the mask bits */
   mask = MASK_DM2_SOUND_EMPTY;
@@ -1680,7 +1780,8 @@ void do_dm2_sound_message_write_bin(node* n, BB_t* m)
   }
 }
 
-void do_dm2_playerinfo_message_write_bin(node* n, BB_t* m)
+void
+do_dm2_playerinfo_message_write_bin(node* n, BB_t* m, unicast_t* u)
 {
   node *cn;
   unsigned long mask;
@@ -1691,7 +1792,12 @@ void do_dm2_playerinfo_message_write_bin(node* n, BB_t* m)
   */
 
   /* at first: the message id */
-  WriteByte(m,node_token_id(n->type));
+  if (u->use == UC_THIS) {
+    WriteByte(m,node_token_id(n->type) | MASK_UNICAST);
+    WriteByte(m,u->client);
+  }
+  else
+    WriteByte(m,node_token_id(n->type));
 #define aaa(x) fprintf(stderr,"%d %d %s\n", x, cn->type, node_token_string(cn->type)); 
 #define bbb(x) fprintf(stderr,"%d mask = 0x%0x\n", x, mask);
   
@@ -1822,7 +1928,8 @@ void do_dm2_playerinfo_message_write_bin(node* n, BB_t* m)
   }
 }
 
-int do_dm2_packetmask_write_bin(node* n, BB_t* m)
+int
+do_dm2_packetmask_write_bin(node* n, BB_t* m)
 {
   node* cn;
   unsigned long mask;
@@ -1957,7 +2064,8 @@ int do_dm2_packetmask_write_bin(node* n, BB_t* m)
 #define aaa(a)
 /* #define aaa(a) { fprintf(stderr,"%d ", a); } */
 
-void do_dm2_packetdata_write_bin(node* n, BB_t* m, unsigned long mask)
+void
+do_dm2_packetdata_write_bin(node* n, BB_t* m, unsigned long mask)
 {
   node* cn;
 
@@ -2055,7 +2163,8 @@ void do_dm2_packetdata_write_bin(node* n, BB_t* m, unsigned long mask)
   }
 }
 
-void do_dm2_spawnbaseline_message_write_bin(node* n, BB_t* m)
+void
+do_dm2_spawnbaseline_message_write_bin(node* n, BB_t* m, unicast_t* u)
 {
   unsigned long mask;
   
@@ -2065,13 +2174,19 @@ void do_dm2_spawnbaseline_message_write_bin(node* n, BB_t* m)
   */
 
   /* at first: the message id */
-  WriteByte(m,node_token_id(n->type));
+  if (u->use == UC_THIS) {
+    WriteByte(m,node_token_id(n->type) | MASK_UNICAST);
+    WriteByte(m,u->client);
+  }
+  else
+    WriteByte(m,node_token_id(n->type));
   
   mask = do_dm2_packetmask_write_bin(n, m);
   do_dm2_packetdata_write_bin(n,m,mask);
 }
 
-void do_dm2_packetentities_message_write_bin(node* n, BB_t* m)
+void
+do_dm2_packetentities_message_write_bin(node* n, BB_t* m, unicast_t* u)
 {
   node* cn;
   unsigned long mask;
@@ -2082,7 +2197,12 @@ void do_dm2_packetentities_message_write_bin(node* n, BB_t* m)
   */
 
   /* at first: the message id */
-  WriteByte(m,node_token_id(n->type));
+  if (u->use == UC_THIS) {
+    WriteByte(m,node_token_id(n->type) | MASK_UNICAST);
+    WriteByte(m,u->client);
+  }
+  else
+    WriteByte(m,node_token_id(n->type));
 
   for (cn = n->down ; cn != NULL ; cn = cn->next ) {
     mask = do_dm2_packetmask_write_bin(cn,m);
@@ -2093,7 +2213,8 @@ void do_dm2_packetentities_message_write_bin(node* n, BB_t* m)
   WriteByte(m,0); /* empty entity as the end mark */
 }
 
-void do_dm2_frame_message_write_bin(node* n, BB_t* m)
+void
+do_dm2_frame_message_write_bin(node* n, BB_t* m, unicast_t* u)
 {
   node *cn;
 
@@ -2103,7 +2224,12 @@ void do_dm2_frame_message_write_bin(node* n, BB_t* m)
   */
 
   /* at first: the message id */
-  WriteByte(m,node_token_id(n->type));
+  if (u->use == UC_THIS) {
+    WriteByte(m,node_token_id(n->type) | MASK_UNICAST);
+    WriteByte(m,u->client);
+  }
+  else
+    WriteByte(m,node_token_id(n->type));
 
   cn = n->down;
 
@@ -2116,6 +2242,7 @@ void do_dm2_frame_message_write_bin(node* n, BB_t* m)
   if (cn->type == TOKEN_UK_B1) {
     do_simple_argument_write_bin(cn->down, m); cn=cn->next;
   }
+  if (cn == NULL) return;
   /* areas */
   if (cn->type == TOKEN_AREAS) {
     /* count at the beginning */
@@ -2131,6 +2258,22 @@ void do_dm2_frame_message_write_bin(node* n, BB_t* m)
   else { /* no areas entry at all */
     WriteByte(m, 0);
   }
+  if (cn == NULL) return;
+  /* connected */
+  if (cn->type == TOKEN_CONNECTED) {
+    /* count at the beginning */
+    {
+      node* c = cn->down;
+      int i;
+      
+      for (i=0 ; c!=NULL ; c=c->next, i++) ;
+      WriteByte(m, i);
+    }
+    do_simple_arguments_write_bin(cn->down, m); cn=cn->next;
+  }
+  else { /* no connected entry at all */
+    WriteByte(m, 0);
+  }
 }  
 
 
@@ -2168,7 +2311,8 @@ void DM2_block_edit(node *n)
         /* ni->down is V_STRING, ni->down->down is the string itself */
         if ((strcmp ((char*)(ni->down->down),"reconnect\n"         )==0) ||
             (strncmp((char*)(ni->down->down),"cmd configstrings",17)==0) ||
-            (strncmp((char*)(ni->down->down),"cmd baselines"    ,13)==0)) {
+            (strncmp((char*)(ni->down->down),"cmd baselines"    ,13)==0) ||
+            (strncmp((char*)(ni->down->down),"record "          , 7)==0)) {
           /* the whole ni must disappear */
           /* unlink this node, correct the start */
           n->down = node_unlink(n->down,ni);
@@ -2196,7 +2340,8 @@ void DM2_block_edit(node *n)
         */
         if ((*(nt = (long*)(ni->down->next->next->down->down))) == 0) *nt = 1;
       }
-      if (ni->type == TOKEN_RECONNECT) {
+      if (ni->type == TOKEN_RECONNECT ||
+          ni->type == TOKEN_DOWNLOAD) {
         /* the whole ni must disappear */
         /* unlink this node, correct the start */
         n->down = node_unlink(n->down,ni);
